#include <zasm/linker.h>
#include <zasm/parser.h>
#include <stdio.h>

static struct linker_t *lnk = NULL;
static struct stream_t *output = NULL;
static struct token_t global_name;

static void write_dump(char *filename, struct list_t *labels, int mode)
{
  FILE *file;
  struct token_t t;
  
  if (!filename)
    return;
  if (!(file = fopen(filename, "w")))
    ERROR("can not open \"%s\" for writing\n", filename);
  foreach(t, labels)
    {
      switch (t.type)
	{
	case TOK_NAME:
	  fprintf(file, "%s ", t.value.t_char_p);
	  string_free(t.value.t_char_p);
	  break;
	case TOK_EQU:
	  fprintf(file, " = ");
	  break;
	case TOK_DWORD_INT_P:
	  if (mode == 0)
	    fprintf(file, " %04hX", *(t.value.t_int_p));
	  else
	    fprintf(file, " 0%xh", *(t.value.t_int_p));
	  break;
	case TOK_NEWLINE:
	  fprintf(file, "\n");
	  break;
	default:
	  WARNING("unknown exported token\n");
	}
    }
  if (fclose(file) < 0)
    ERROR("I/O error\n");
}

static void dump_label(struct list_t *list, struct token_t name)
{
  struct expr_t *expr;
  if (!list)
    return;
  list_add_end(list, token_copy(name));
  expr = expr_new(&(lnk->expr_alloc));
  expr->type = EXPR_NAME;
  expr->value.name.name = string_copy(name.value.t_char_p);
  expr->value.name.part = NULL;
  name.type = TOK_EQU;
  list_add_end(list, name);
  name.type = TOK_DWORD_INT_P;
  name.value.t_int_p = linker_get_int(lnk);
  link_namespace_lookup(lnk->ns, expr, lnk->address, lnk->page, lnk->instr_start, name.value.t_int_p);  
  list_add_end(list, name);
  name.type = TOK_NEWLINE;
  list_add_end(list, name);  
}


static inline struct token_t pc_parser_0(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t result;
    result.type = TOK_EXPR;
    result.value.t_struct_expr_t_p = parse_expr(input, token, &(lnk->expr_alloc));
    return result;
  
}
static inline 
struct token_t expr_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_0(input,token);
}

static inline struct token_t pc_parser_1(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,result;
    lnk->address+=1;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
result.position = token->position;
    result.type = TOK_BYTE_INT_P;
    result.value.t_int_p = linker_get_int(lnk);
    link_namespace_lookup(lnk->ns, expr.value.t_struct_expr_t_p, lnk->address, lnk->page, lnk->instr_start, result.value.t_int_p);
    return result;
  
}
static inline 
struct token_t byte_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_1(input,token);
}

static inline struct token_t pc_parser_2(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,result;
    lnk->address+=2;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
result.position = token->position;
    result.type = TOK_WORD_INT_P;
    result.value.t_int_p = linker_get_int(lnk);
    link_namespace_lookup(lnk->ns, expr.value.t_struct_expr_t_p, lnk->address, lnk->page, lnk->instr_start, result.value.t_int_p);
    return result;
  
}
static inline 
struct token_t word_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_2(input,token);
}

static inline struct token_t pc_parser_3(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,result;
    lnk->address+=4;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
result.position = token->position;
    result.type = TOK_DWORD_INT_P;
    result.value.t_int_p = linker_get_int(lnk);
    link_namespace_lookup(lnk->ns, expr.value.t_struct_expr_t_p, lnk->address, lnk->page, lnk->instr_start, result.value.t_int_p);
    return result;
  
}
static inline 
struct token_t dword_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_3(input,token);
}

static inline struct token_t pc_parser_4(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
lnk->address = expr_calculate(expr.value.t_struct_expr_t_p);
    expr.type = TOK_SET_PAGE;
    expr.position = token->position;
    expr.value.t_struct_page_addr_t.addr = lnk->address;
    expr.value.t_struct_page_addr_t.page = lnk->page;
    return expr;
  
}
static inline 
struct token_t org_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_4(input,token);
}

static inline struct token_t pc_parser_5(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
lnk->page = expr_calculate(expr.value.t_struct_expr_t_p);
    expr.position = token->position;
    expr.type = TOK_SET_PAGE;
    expr.value.t_struct_page_addr_t.addr = lnk->address;
    expr.value.t_struct_page_addr_t.page = lnk->page;
    return expr;
  
}
static inline 
struct token_t page_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_5(input,token);
}

static inline struct token_t pc_parser_6(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr;
      expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
lnk->page_start = expr_calculate(expr.value.t_struct_expr_t_p);
    expr.type = TOK_PC_SUCCESS;
    return expr;
  
}
static inline 
struct token_t page_start_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_6(input,token);
}

static inline struct token_t pc_parser_7(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    lnk->page++;
    lnk->address = lnk->page_start;
    pc_retval.position = token->position;
    pc_retval.type = TOK_SET_PAGE;
    pc_retval.value.t_struct_page_addr_t.page = lnk->page;
    pc_retval.value.t_struct_page_addr_t.addr = lnk->address;
    return pc_retval;
  
}
static inline 
struct token_t newpage_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_7(input,token);
}

static inline struct token_t pc_parser_8(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    link_namespace_add_level(lnk->ns);
    pc_retval.type = TOK_PC_SUCCESS;
    return pc_retval;
  
}
static inline 
struct token_t begin_section_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_8(input,token);
}

static inline struct token_t pc_parser_9(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    link_namespace_del_level(lnk->ns);
    pc_retval.type = TOK_PC_SUCCESS;
    return pc_retval;    
  
}
static inline 
struct token_t end_section_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_9(input,token);
}

static inline struct token_t pc_parser_10(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    lnk->instr_start = lnk->address;
    pc_retval.type = TOK_PC_SUCCESS;
    return pc_retval;    
  
}
static inline 
struct token_t begin_instr_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_10(input,token);
}

static inline struct token_t pc_parser_11(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_12(struct feeder_t *input, struct token_t *token,int type)
{
  struct token_t result;
  if (token->type == type)
  {
    result = *token;
    *token = feeder_get(input);
  }
  else result.type = TOK_PC_FAIL1;
  return result;
}

static inline struct token_t pc_parser_13(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_14(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t name, expr;
      pc_retval=pc_parser_11(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  name=pc_parser_12(input,token,TOK_NAME);
  if ((name.type == TOK_PC_FAIL)||(name.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      name.type = TOK_PC_FAIL;
    return name;
  }
  pc_called = 1;
  pc_retval=pc_parser_13(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
dump_label(lnk->dump_labels, name);
    link_namespace_add(lnk->ns, name.value.t_char_p, expr.value.t_struct_expr_t_p, lnk->address, lnk->page, lnk->instr_start);
    token_free(name);
    name.type = TOK_PC_SUCCESS;
    return name;
  
}
static inline 
struct token_t equ_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_14(input,token);
}

static inline struct token_t pc_parser_15(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_16(struct feeder_t *input, struct token_t *token,int type)
{
  struct token_t result;
  if (token->type == type)
  {
    result = *token;
    *token = feeder_get(input);
  }
  else result.type = TOK_PC_FAIL1;
  return result;
}

static inline struct token_t pc_parser_17(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t name;
      pc_retval=pc_parser_15(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  name=pc_parser_16(input,token,TOK_NAME);
  if ((name.type == TOK_PC_FAIL)||(name.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      name.type = TOK_PC_FAIL;
    return name;
  }
  pc_called = 1;
dump_label(lnk->export_labels, name);
    link_namespace_export(lnk->ns, name.value.t_char_p);
    token_free(name);
    name.type = TOK_PC_SUCCESS;
    return name;
  
}
static inline 
struct token_t export_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_17(input,token);
}

static inline struct token_t pc_parser_18(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_19(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,temp;
    int dest;
      pc_retval=pc_parser_18(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
dest = expr_calculate(expr.value.t_struct_expr_t_p);
    if (dest < lnk->address)
      T_ERROR(*token, "trying to align at a past address (%i), current is %i\n", dest, lnk->address);
    temp.type = TOK_BYTE_INT;
    temp.position = token->position;
    temp.value.t_int = 0xFF;
    while (dest > lnk->address)
    {
      lnk->address++;
      stream_put(output, temp);
    }
    expr.type = TOK_PC_SUCCESS;
    return expr;
  
}
static inline 
struct token_t align_addr_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_19(input,token);
}

static inline struct token_t pc_parser_20(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_21(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,temp;
    int dest,mult,t;
      pc_retval=pc_parser_20(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
mult = expr_calculate(expr.value.t_struct_expr_t_p);
    t = lnk->address % mult;
    dest = lnk->address+(mult-(t?t:mult));
    temp.type = TOK_BYTE_INT;
    temp.position = token->position;
    temp.value.t_int = 0xff;
    while (dest > lnk->address)
    {
      lnk->address++;
      stream_put(output, temp);
    }
    expr.type = TOK_PC_SUCCESS;
    return expr;
  
}
static inline 
struct token_t align_mult_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_21(input,token);
}

static inline struct token_t pc_parser_22(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_23(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t name,expr;
      pc_retval=pc_parser_22(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
dump_label(lnk->dump_labels, global_name);
    link_namespace_add(lnk->ns, global_name.value.t_char_p, expr.value.t_struct_expr_t_p, lnk->address, lnk->page, lnk->instr_start);
    token_free(global_name);
    name.type = TOK_PC_SUCCESS;
    return name;
  
}
static inline 
struct token_t name_equ_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_23(input,token);
}

static inline struct token_t pc_parser_24(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t name;
    struct expr_t *expr;

    expr = expr_new(&(lnk->expr_alloc));
    expr->type = EXPR_NAME;
    expr->value.name.name = string_new("$");
    expr->value.name.part = string_new("end");
    dump_label(lnk->dump_labels, global_name);
    link_namespace_add(lnk->ns, global_name.value.t_char_p, expr, lnk->address, lnk->page, lnk->instr_start);
    token_free(global_name);
    name.type = TOK_PC_SUCCESS;
    return name;
  
}
static inline 
struct token_t label_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_24(input,token);
}

static inline struct token_t pc_parser_25(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t name;
    dump_label(lnk->export_labels, global_name);
    link_namespace_export(lnk->ns, global_name.value.t_char_p);
    token_free(global_name);
    name.type = TOK_PC_SUCCESS;
    return name;
  
}
static inline 
struct token_t name_export_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_25(input,token);
}

static inline struct token_t pc_parser_26(struct feeder_t *input, struct token_t *token,int type)
{
  struct token_t result;
  if (token->type == type)
  {
    result = *token;
    *token = feeder_get(input);
  }
  else result.type = TOK_PC_FAIL1;
  return result;
}

static inline struct token_t pc_parser_27(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_28(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  switch (token->type)
  {
  case TOK_EQU:
    *token = feeder_get(input);
    result = name_equ_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
			      TOK_OP_EQU:
    *token = feeder_get(input);
    result = name_equ_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
			      TOK_EXPORT:
    *token = feeder_get(input);
    result = name_export_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  default:
    result.type = TOK_PC_FAIL1;
    return result;
  }
}

static inline struct token_t pc_parser_29(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  result=pc_parser_28(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;  result=label_p(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;
  return result;
}

static inline struct token_t pc_parser_30(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;
  global_name=pc_parser_26(input,token,TOK_NAME);
  if ((global_name.type == TOK_PC_FAIL)||(global_name.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      global_name.type = TOK_PC_FAIL;
    return global_name;
  }
  pc_called = 1;
  pc_retval=pc_parser_27(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  pc_retval=pc_parser_29(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
return pc_retval;

}
static inline 
struct token_t name_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_30(input,token);
}

static inline struct token_t pc_parser_31(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t result;
    if (token->type == TOK_BYTE_INT)
      lnk->address+=1;
    else if (token->type == TOK_WORD_INT)
      lnk->address+=2;
    else if (token->type == TOK_DWORD_INT)
      lnk->address+=4;
    else
    {
      result.type = TOK_PC_FAIL1;
      return result;
    }
    result = *token;
    *token = feeder_get(input);
    return result;
  
}
static inline 
struct token_t int_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_31(input,token);
}

static inline struct token_t pc_parser_32(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_33(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t result;
      pc_retval=pc_parser_32(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
if (token->type == TOK_NAME)
    {
      T_WARNING(*token, "ignoring unknown directive \"%s\"\n", token->value.t_char_p);
      *token = feeder_get(input);
    }
    result.type = TOK_PC_SUCCESS;
    return result;
  
}
static inline 
struct token_t ignore_unknown_directive_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_33(input,token);
}

static inline struct token_t pc_parser_34(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_35(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,temp;
    int dest;
      pc_retval=pc_parser_34(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
dest = lnk->address+expr_calculate(expr.value.t_struct_expr_t_p);
    if (dest < lnk->address)
      T_ERROR(*token, "trying to skip a negative amount of bytes\n");
    temp.type = TOK_BYTE_INT;
    temp.position = token->position;
    temp.value.t_int = 0xFF;
    while (dest > lnk->address)
    {
      lnk->address++;
      stream_put(output, temp);
    }
    expr.type = TOK_PC_SUCCESS;
    return expr;
  
}
static inline 
struct token_t block_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_35(input,token);
}

static inline struct token_t pc_parser_36(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_37(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  while (token->type == TOK_SPACE)
    *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}

static inline struct token_t pc_parser_38(struct feeder_t *input, struct token_t *token){
  struct token_t pc_retval;
  int pc_called = 0;

    struct token_t expr,temp;
    int dest;
      pc_retval=pc_parser_36(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
  expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
  pc_retval=pc_parser_37(input,token);
  if ((pc_retval.type == TOK_PC_FAIL)||(pc_retval.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      pc_retval.type = TOK_PC_FAIL;
    return pc_retval;
  }
  pc_called = 1;
dest = lnk->address+expr_calculate(expr.value.t_struct_expr_t_p);
    temp.type = TOK_BYTE_INT;
    temp.position = token->position;
    temp.value.t_int = 0xFF;
    if (token->type == TOK_COMMA)
    {
      *token = feeder_get(input);
        expr=expr_p(input,token);
  if ((expr.type == TOK_PC_FAIL)||(expr.type == TOK_PC_FAIL1))
  {
    if (pc_called)
      expr.type = TOK_PC_FAIL;
    return expr;
  }
  pc_called = 1;
temp.value.t_int = expr_calculate(expr.value.t_struct_expr_t_p);
    }
    if (dest < lnk->address)
      T_ERROR(*token, "trying to fill a negative amount of bytes\n");
    while (dest > lnk->address)
    {
      lnk->address++;
      stream_put(output, temp);
    }
    expr.type = TOK_PC_SUCCESS;
    return expr;
  
}
static inline 
struct token_t fill_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_38(input,token);
}

static inline struct token_t pc_parser_39(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  switch (token->type)
  {
  case TOK_BYTE:
    *token = feeder_get(input);
    result = byte_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_WORD:
    *token = feeder_get(input);
    result = word_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_DWORD:
    *token = feeder_get(input);
    result = dword_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_ALIGN_ADDR:
    *token = feeder_get(input);
    result = align_addr_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_ALIGN_MULT:
    *token = feeder_get(input);
    result = align_mult_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_ORG:
    *token = feeder_get(input);
    result = org_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_PAGE:
    *token = feeder_get(input);
    result = page_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_PAGE_START:
    *token = feeder_get(input);
    result = page_start_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_NEWPAGE:
    *token = feeder_get(input);
    result = newpage_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_EQU:
    *token = feeder_get(input);
    result = equ_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_BEGIN_SECTION:
    *token = feeder_get(input);
    result = begin_section_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_END_SECTION:
    *token = feeder_get(input);
    result = end_section_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_EXPORT:
    *token = feeder_get(input);
    result = export_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_BEGIN_INSTR:
    *token = feeder_get(input);
    result = begin_instr_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_DOT:
    *token = feeder_get(input);
    result = ignore_unknown_directive_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_SHARP:
    *token = feeder_get(input);
    result = ignore_unknown_directive_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_FILL:
    *token = feeder_get(input);
    result = fill_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  case 
				    TOK_BLOCK:
    *token = feeder_get(input);
    result = block_p(input,token);
    if (result.type == TOK_PC_FAIL1)
      result.type = TOK_PC_FAIL;
    return result;
  default:
    result.type = TOK_PC_FAIL1;
    return result;
  }
}

static inline struct token_t pc_parser_40(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  result = *token;
  *token = feeder_get(input);
  return result;
}

static inline struct token_t pc_parser_41(struct feeder_t *input, struct token_t *token)
{
  struct token_t result;
  result=pc_parser_39(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;  result=name_p(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;  result=int_p(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;  result=pc_parser_40(input,token);
  if (result.type != TOK_PC_FAIL1)
    return result;
  return result;
}

struct token_t linker_p(struct feeder_t *input, struct token_t *token)
{
  return pc_parser_41(input,token);
}













































void linker_execute(struct feeder_t *input, struct stream_t *outputS, struct linker_t *linker)
{
  struct token_t result;
  
  output = outputS;
  lnk = linker;
  if (lnk->look_ahead.type == TOKENS_COUNT)
    lnk->look_ahead = feeder_get(input);
  while (1)
    {
      if (lnk->look_ahead.type == TOK_END_OF_FILE)
        break;
      result = linker_p(input, &(lnk->look_ahead));
      if (result.type == TOK_PC_SUCCESS)
        continue;
      if (result.type == TOK_LIST)
        list_add_list_end(&(output->list), result.value.t_struct_list_t_p);
      else if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))
        T_ERROR(lnk->look_ahead, "syntax error\n");
      else
        stream_put(output, result);
    }
  if (linker->expr_alloc)
    expr_alloc_reset(linker->expr_alloc);
  if (lnk->look_ahead.type == TOK_END_OF_FILE)
    {
      stream_put(output, lnk->look_ahead);
      write_dump(lnk->dump_file, lnk->dump_labels, 0);
      write_dump(lnk->export_file, lnk->export_labels, 1);
      linker_free(linker);
      feeder_free(input);
    }
}
