#define LINK_LOOKUP_C 1

#include <zasm/link_lookup.h>
#include <zasm/link_expr.h>

ALLOC_DEFINE(link_lookup, ALLOC_DEF_COUNT);

struct link_lookup_t *link_lookup_new(struct link_level_t *level, struct expr_t *expr, int addr, int page, int instr_start, int *dest)
{
  struct link_lookup_t *val;

  val = link_lookup_alloc_get(&(level->lookup_alloc));
  link_lookup_fill(val, level, expr, addr, page, instr_start, dest);
  return val;
}

void link_lookup_fill(struct link_lookup_t *lookup, struct link_level_t *level,
                      struct expr_t *expr, int addr, int page, int instr_start, int *dest)
{
  lookup->dest = dest;
  lookup->expr = link_expr_from_expr(level, expr, addr, page, instr_start);
  lookup->next = level->exprs;
  level->exprs = lookup;
}

struct link_lookup_t *link_lookup_new_copy(struct link_level_t *level, struct link_lookup_t *orig)
{
  struct link_lookup_t *val;

  val = link_lookup_alloc_get(&(level->lookup_alloc));
  val->dest = orig->dest;
  val->expr = link_expr_copy(orig->expr, level);
  val->next = level->exprs;
  level->exprs = val;
  return val;
} 

void link_lookup_error_unresolved(struct link_lookup_t *lookup)
{
  for (; lookup; lookup = lookup->next)
    {
      link_expr_simplify(lookup->expr);
      link_expr_print_unresolved(lookup->expr);
    }
  ERROR("there were unresolved labels\n");
}

void link_lookup_copy(struct link_lookup_t *lookup, struct link_level_t *dest)
{
  struct link_lookup_t *new;
  
  for (; lookup; lookup = lookup->next)
    {
      link_expr_simplify(lookup->expr);
      if (lookup->expr->type == EXPR_NUMBER)
	*(lookup->dest) = lookup->expr->value.value;
      else if (!dest)
	link_lookup_error_unresolved(lookup);
      else
	new = link_lookup_new_copy(dest, lookup);
    }
}

void link_lookup_delete(struct link_lookup_t *lookup)
{}

void link_lookup_free(struct link_lookup_t *lookup)
{}
